/* SPDX-License-Identifier: GPL-2.0-or-later */
/*
 * Copyright (C) 2014 Red Hat, Inc.
 */

#include "src/core/nm-default-daemon.h"

#include "nm-bt-error.h"

NM_CACHED_QUARK_FCN("nm-bt-error", nm_bt_error_quark);
