/* SPDX-License-Identifier: LGPL-2.1-or-later */
/*
 * Copyright (C) 2015 Red Hat, Inc.
 */

#ifndef __NM_DEFAULT_DAEMON_H__
#define __NM_DEFAULT_DAEMON_H__

/*****************************************************************************/

#include "libnm-glib-aux/nm-default-glib-i18n-prog.h"

#undef NETWORKMANAGER_COMPILATION
#define NETWORKMANAGER_COMPILATION NM_NETWORKMANAGER_COMPILATION_DAEMON

/*****************************************************************************/

#include "nm-version.h"

#include "nm-core-types.h"
#include "nm-errors.h"
#include "nm-types.h"
#include "libnm-log-core/nm-logging.h"

/*****************************************************************************/

#endif /* __NM_DEFAULT_DAEMON_H__ */
