/* SPDX-License-Identifier: LGPL-2.1-or-later */
/*
 * Copyright (C) 2015 Red Hat, Inc.
 */

#ifndef __NM_DEFAULT_SYSTEMD_H__
#define __NM_DEFAULT_SYSTEMD_H__

/*****************************************************************************/

#include "libnm-glib-aux/nm-default-glib.h"

#undef NETWORKMANAGER_COMPILATION
#define NETWORKMANAGER_COMPILATION NM_NETWORKMANAGER_COMPILATION_SYSTEMD

/*****************************************************************************/

#endif /* __NM_DEFAULT_SYSTEMD_H__ */
